#!/bin/bash
#
### BEGIN INIT INFO
# Provides:         ptdbnssd
# Required-Start:   $network
# Required-Stop:    $network
# X-Start-Before:   openafs-client
# Default-Start:    S
# Default-Stop:     0 6
# Short-Description: Daemon for PTDB-based NSS-plugin
# Description:      This daemon takes care of communicating with the
#                   AFS cell's PTDB and answering the NSS-using processes'
#                   name-resolving requests.
### END INIT INFO
#
# -- Frank Burkhardt <burk@cbs.mpg.de>  Sun, 08 Nov 2009 21:33:04 +0100


PATH=/sbin:/bin:/usr/sbin:/usr/bin
DAEMON=/usr/bin/ptdbnssd
NAME=ptdbnssd
DESC=AFS-name-service

test -x $DAEMON || exit 0

[ -e /etc/default/$NAME ] && source /etc/default/$NAME

CHILDS=3
HOMEDIRS=auto
SHELLS=bash

DAEMON_OPTS="-c $CHILDS -H $HOMEDIRS -S $SHELLS"

case "$1" in
  start)
	echo -n "Starting $DESC: "
	start-stop-daemon --start --quiet --pidfile /var/run/$NAME.pid \
		--background --chuid nobody:nogroup --exec $DAEMON -- $DAEMON_OPTS
	echo "$NAME."
	;;
  stop)
	echo -n "Stopping $DESC: "
	killall ptdbnssd || true
	echo "$NAME."
	;;
  restart|force-reload)
	echo -n "Restarting $DESC: "
	start-stop-daemon --stop --quiet --pidfile \
		/var/run/$NAME.pid --exec $DAEMON
	sleep 1
	start-stop-daemon --start --quiet --pidfile \
		/var/run/$NAME.pid --background --chuid nobody:nogroup --exec $DAEMON -- $DAEMON_OPTS
	echo "$NAME."
	;;
  *)
	N=/etc/init.d/$NAME
	echo "Usage: $N {start|stop|restart|force-reload}" >&2
	exit 1
	;;
esac

exit 0
