#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <errno.h>
#include <pwd.h>
#include <nss.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <string.h>
#include "ptdbnss.h"

extern int ptsid2name(int uid,char **buffer,size_t *buflen,struct ptdbconf_t *config);
extern int ptsname2id(char *name,uid_t *uid,char **buffer,struct ptdbconf_t *config);
extern int cpstr( char *str, char **buf, size_t *buflen);



char pathprefix[300];
char cellname[256];
int pathprefixlen=0;


/*
	result=read_cellname();
	
	Reads the name of the local workstation's cell out of /etc/openafs/ThisCell .
	
	result == 1, only if the cell was read successfully.
	
	This function should be thread-safe because /etc/openafs/ThisCell is not considered
	to change often.
*/
int read_cellname() {
	FILE *thiscell;
	int len;
  
	thiscell=fopen("/etc/openafs/ThisCell","r");
	if ( thiscell == NULL ) {
		return 0;
	}
	len=fread(cellname,1,256,thiscell);
	if ( ! feof(thiscell) ) {
		// Cellname too long
		fclose(thiscell);
		strcpy(pathprefix,"/tmp/\0");
		pathprefixlen=5;
		return 0;
	}
	fclose(thiscell);
	if ( cellname[len-1] == '\n' ) {
		len--;
	}
	cellname[len]='\0';
	sprintf(pathprefix,"/afs/%s/user/",cellname);
	pathprefixlen=strlen(pathprefix);
	return 1;
}



/*
	result=get_homedir(char *name,char **buffer,size_t *buflen)
	
	Writes the guessed Homedirectory of a given user 'name' into
	'*buffer', increasing *buflen accordingly.
	
	result == 1, only if the buffer was big enough.
*/
int get_homedir(char *name,char **buffer,size_t *buflen,struct ptdbconf_t *config) {
	char buf[256];
	int temp;
	char *b;
	b=*buffer;
	switch ( config->homedirs_method ) {
		case HOMEDIR_USER_1LEVEL:
			pathprefix[pathprefixlen]=0;
			strncpy(&pathprefix[pathprefixlen],name,40);
			if ( cpstr(pathprefix,buffer,buflen) ) {
				return 1;
			}
			break;
		case HOMEDIR_ADMINLINK:
			if ( snprintf(buf,256,"/afs/%s/admin/homedirs/%s",cellname,name) > 0 ) {
				temp=readlink(buf,*buffer,*buflen);
				if ( temp > -1) {
					b[temp]=0;
					*buflen = *buflen - temp - 1;
					return 1;
				}
			}
			if ( cpstr("/tmp",buffer,buflen) ) {
				return 1;
			}
			break;
		case HOMEDIR_HOME_1LEVEL:
			if ( snprintf(buf,256,"/afs/%s/home/%s",cellname,name) > 0 ) {
				if ( cpstr(buf,buffer,buflen) ) {
					return 1;
				}
			}
			break;
		case HOMEDIR_USEREXT:
			if ( snprintf(buf,256,"/afs/%s/user-ext/%s",cellname,name) > 0 ) {
				if ( cpstr(buf,buffer,buflen) ) {
					return 1;
				}
			}
			break;
		case HOMEDIR_USER_2LEVEL1X:
			if ( snprintf(buf,256,"/afs/%s/user/%.1s/%s",cellname,name,name) > 0 ) {
				if ( cpstr(buf,buffer,buflen) ) {
					return 1;
				}
			}
			break;
		case HOMEDIR_USER_3LEVEL12X:
			if ( snprintf(buf,256,"/afs/%s/user/%.1s/%.2s/%s",cellname,name,name,name) > 0 ) {
				if ( cpstr(buf,buffer,buflen) ) {
					return 1;
				}
			}
			break;
		case HOMEDIR_LOCAL_HOME_MINOR:
			if ( snprintf(buf,256,"/home/%s",name) > 0 ) {
				if ( cpstr(buf,buffer,buflen) ) {
					return 1;
				}
			}
			break;
		case HOMEDIR_LOCAL_HOME_CAPITAL:
			if ( snprintf(buf,256,"/HOME/%s",name) > 0 ) {
				if ( cpstr(buf,buffer,buflen) ) {
					return 1;
				}
			}
			break;
	}
	return 0;
}

int get_shell(char *name,char **buffer,size_t *buflen,struct ptdbconf_t *config) {
	char buf[256];
	int temp;
	char *b;
	b=*buffer;
	switch ( config->shells_method ) {
		case SHELL_BASH:
			if ( cpstr("/bin/bash",buffer, buflen) ) {
				return 1;
			}
			break;
		case SHELL_ADMINLINK:
			if ( snprintf(buf,256,"/afs/%s/admin/shells/%s",cellname,name) > 0 ) {
				temp=readlink(buf,*buffer,*buflen);
				if ( temp > -1) {
					b[temp]=0;
					*buflen = *buflen - temp - 1;
					return 1;
				}
			}
			// Fallback, if no shell is defined by the admin
			if ( cpstr("/bin/bash",buffer,buflen) ) {
				return 1;
			}
			break;
		case SHELL_USERLINK:
			if ( snprintf(buf,256,"/afs/%s/user/%s/.loginshell",cellname,name) > 0 ) {
				temp=readlink(buf,*buffer,*buflen);
				if ( temp > -1) {
					b[temp]=0;
					*buflen = *buflen - temp - 1;
					return 1;
				}
			}
			// Fallback, if no shell is explicitely set by the user
			if ( cpstr("/bin/bash",buffer,buflen) ) {
				return 1;
			}
			break;
		case SHELL_SCPONLY:
			if ( cpstr("/usr/bin/scponly",buffer, buflen) ) {
				return 1;
			}
	}
	return 0;
}





/*
	This is a the ptdb-getpwuid-function.
*/
enum nss_status _nss_ptdb_getpwuid_r (uid_t uid, struct passwd *result_buf,
 char *buffer, size_t buflen, int *errnop ) {
	struct ptdbconf_t config;
	int temp;

	/* Read the cellname */
	if ( pathprefixlen == 0 ) {
		read_cellname();
		if ( pathprefixlen == 0 ) {
			return NSS_STATUS_NOTFOUND;
		}
	}
	result_buf->pw_name = buffer;
	if ( ptsid2name( uid, &buffer, &buflen, &config ) < 1 ) {
		*errnop = ENOENT;
		return NSS_STATUS_NOTFOUND;
	}
#ifdef LIMIT_USERNAME_CHARS
	if ( strlen(result_buf->pw_name) > LIMIT_USERNAME_CHARS ) {
		result_buf->pw_name[LIMIT_USERNAME_CHARS] = '\0';
		buflen = buflen + ( buffer - &result_buf->pw_name[LIMIT_USERNAME_CHARS+1] );
		buffer = &result_buf->pw_name[LIMIT_USERNAME_CHARS+1];
	}
#endif

	/* set the password to "x" */
	result_buf->pw_passwd = buffer;
	if ( ! cpstr("x",&buffer, &buflen) ) {
		/* if ( log ) { fprintf(log,"%d:%d ERANGE, TRYAGAIN on passwd\n",getpid(),uid ); fclose(log); } */
		*errnop = ERANGE;
		return NSS_STATUS_TRYAGAIN;
	}

	/* the uid and gid are both the uid passed in */
	result_buf->pw_uid = uid;
	result_buf->pw_gid = 65534;

	/* make the gecos the same as the PTS name */
	result_buf->pw_gecos = buffer;
	if ( ! cpstr(result_buf->pw_name, &buffer, &buflen ) ) {
		/* if ( log ) { fprintf(log,"%d:%d ERANGE, TRYAGAIN on gecos\n",getpid(), uid ); fclose(log); } */
		*errnop = ERANGE;
		return NSS_STATUS_TRYAGAIN;
	}

	// Set the homedirectory
	result_buf->pw_dir = buffer;
	temp=get_homedir(result_buf->pw_name,&buffer,&buflen,&config);
	if ( ! temp ) {
		*errnop = ERANGE;
		return NSS_STATUS_NOTFOUND;
	}

	// Set the login shell
	result_buf->pw_shell = buffer;
	temp=get_shell(result_buf->pw_name,&buffer,&buflen,&config);
	if ( ! temp ) {
		*errnop = ERANGE;
		return NSS_STATUS_NOTFOUND;
	}

	*errnop = errno;
	return NSS_STATUS_SUCCESS;
}



/*
	This is the ptdb-getpwnam-function.
*/
enum nss_status _nss_ptdb_getpwnam_r (char *name, struct passwd *result_buf,
 char *buffer, size_t buflen, int *errnop ) {
	uid_t uid;
	int res;
	struct ptdbconf_t config;

	res=ptsname2id(name,&uid,&buffer,&config);

	if ( res != 1 ) {
		return NSS_STATUS_NOTFOUND;
	}
	if ( uid == 32766 ) {
		if ( strcmp(name,"anonymou") != 0 ) {
			return NSS_STATUS_NOTFOUND;
		}
	}
	if ( uid != -1 ) {
		// This causes an additional PTDB-lookup and should be removed in the future
		return _nss_ptdb_getpwuid_r (uid, result_buf,buffer,buflen, errnop );
	} else {
		*errnop = ENOENT;
		return NSS_STATUS_NOTFOUND;
	}
}

