#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <errno.h>
#include <pwd.h>
#include <nss.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <sys/select.h>
#include <unistd.h>
#include <stdint.h>
#include "ptdbnss.h"
#include <string.h>

extern int cpstr( char *str, char **buf, size_t *buflen);

int ptdbnssd_timeout=TIMEOUT+2;

void copy_conf(struct ptdbconf_t *from,struct ptdbconf_t *to) {
	to->homedirs_method=from->homedirs_method;
	to->shells_method=from->shells_method;
}

int ptsid2name(int uid,char **buffer,size_t *buflen,struct ptdbconf_t *config) {
	int udp_socket;
	struct sockaddr_in laddr,raddr;
	ptdbnsspacket_t sendbuffer;
	int res;
	int running=1;
	int *magicp;
	int magic;

	fd_set readfds;
	struct timeval timeout;

	magicp=(void *)buffer;
	magic=(uint16_t)*magicp;

	raddr.sin_family=AF_INET;
	raddr.sin_port=htons(PTDBNSSD_PORT);
	raddr.sin_addr.s_addr=htonl(INADDR_LOOPBACK); // localhost

	laddr.sin_family=AF_INET;
	laddr.sin_port=htons(0);
	laddr.sin_addr.s_addr=htonl(INADDR_LOOPBACK); // localhost
  
	udp_socket = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if ( udp_socket == -1 ) {
		return 0;
	}
	sendbuffer.header.ptype=NAME_REQUEST;
	sendbuffer.header.id=magic;   // a "random" number
	sendbuffer.header.version=HEADERVERSION;
	sendbuffer.data.uid=uid;
	res=bind(udp_socket,(struct sockaddr*)&laddr,sizeof(laddr));
	if ( res == -1 ) {
		close(udp_socket);
		return 0;
	}
	res=sendto(udp_socket,&sendbuffer,sizeof(struct ptdbnssheader_t)+sizeof(uid),0,(struct sockaddr*)&raddr,sizeof(raddr));
	if ( res == -1 ) {
		close(udp_socket);
		return 0;
	}
  
	while ( running ) {
		running=0;
		FD_ZERO(&readfds);
		FD_SET(udp_socket,&readfds);
		timeout.tv_sec=10;
		timeout.tv_usec=0;
		res=select(udp_socket+1,&readfds,NULL,NULL,&timeout);
		if ( res == 1 ) {
			res=recv(udp_socket,&sendbuffer,sizeof(ptdbnsspacket_t),0);
			if ( res >= 0 ) {
				if ( ( res >= sizeof(struct ptdbnssheader_t) ) && ( sendbuffer.header.ptype == NAME_REPLY ) && ( sendbuffer.header.id == magic ) ) {
					cpstr(sendbuffer.data.name,buffer,buflen);
					ptdbnssd_timeout=TIMEOUT+2;
					close(udp_socket);
					if ( sendbuffer.header.version != HEADERVERSION ) {
						// Unknown header version, ignoring this answer
						return -2;
					}
					if ( sendbuffer.data.name[0] == 0 ) {
						// zero name returned -> unknown user
						return 0;
					} else {
						copy_conf(&sendbuffer.header.conf,config);
						return 1;
					}
				} else {
					// We received a wrong packet, da capo...
					running=1;
				}
			} else {
				// error receiving...
				close(udp_socket);
				return -1;
			}
		} else {
			if ( ( res == -1 ) && ( errno == EINTR ) ) {
				// we were Interrupted by a signal, da capo...
				running=1;
			}
		}
	}
	// no response -> reduce timeout to prevent applications from locking
	ptdbnssd_timeout=1;
	close(udp_socket);
	return -1;
}




int ptsname2id(char *name,uid_t *uid,char **buffer,struct ptdbconf_t *config) {
	int udp_socket;
	struct sockaddr_in laddr,raddr;
	ptdbnsspacket_t sendbuffer;
	int res;
	int running=1;
	int *magicp;
	int magic;

	fd_set readfds;
	struct timeval timeout;

	/* We take a 'random' number (adress if the buffer in memory) which has to be sent back
		to the client in order to accept the result */
	magicp=(void *)buffer;
	magic=(uint16_t)*magicp;
  
	raddr.sin_family=AF_INET;
	raddr.sin_port=htons(PTDBNSSD_PORT);
	raddr.sin_addr.s_addr=htonl(INADDR_LOOPBACK); // localhost

	laddr.sin_family=AF_INET;
	laddr.sin_port=htons(0);
	laddr.sin_addr.s_addr=htonl(INADDR_LOOPBACK); // localhost
  
	udp_socket = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if ( udp_socket == -1 ) {
		return 0;
	}
	sendbuffer.header.ptype=UID_REQUEST;
	sendbuffer.header.id=magic;
	sendbuffer.header.version=HEADERVERSION;
	strcpy(sendbuffer.data.name,name);
	
	res=bind(udp_socket,(struct sockaddr*)&laddr,sizeof(laddr));
	if ( res == -1 ) {
		close(udp_socket);
		return 0;
	}
	res=sendto(udp_socket,&sendbuffer,sizeof(struct ptdbnssheader_t)+1+strlen(name),0,(struct sockaddr*)&raddr,sizeof(raddr));
	if ( res == -1 ) {
		close(udp_socket);
		return 0;
	}
	
	while ( running ) {
		running=0;
		FD_ZERO(&readfds);
		FD_SET(udp_socket,&readfds);
		timeout.tv_sec=10;
		timeout.tv_usec=0;
		res=select(udp_socket+1,&readfds,NULL,NULL,&timeout);
		if ( res == 1 ) {
			res=recv(udp_socket,&sendbuffer,sizeof(ptdbnsspacket_t),0);
			if ( res >= 0 ) {
				if ( ( res >= sizeof(struct ptdbnssheader_t) ) && ( sendbuffer.header.ptype == UID_REPLY ) && ( sendbuffer.header.id == magic ) ) {
					ptdbnssd_timeout=TIMEOUT;
					close(udp_socket);
					if ( sendbuffer.header.version != HEADERVERSION ) {
						// Unknown header version, ignoring this answer
						return -2;
					}
					if ( sendbuffer.data.uid == 0 ) {
						// User not found in PTDB
						return 0;
					} else {
						// User found in PTDB
						*uid=sendbuffer.data.uid;
						copy_conf(&sendbuffer.header.conf,config);
						return 1;
					}
				} else {
					// we received a malformed packet, da capo..
					running=1;
				}
			} else {
				// error receiving from UDP-socket
				close(udp_socket);
				return -1;
			}
		} else {
			if ( ( res == -1 ) && ( errno == EINTR ) ) {
				// we were Interrupted by a signal. da capo...
				running=1;
			}
		}
	}
	// no response, we'll set timeout to 1sec, preventing current program to lock
	if ( ptdbnssd_timeout != 1 ) {
		printf("W: libnss-ptdb: unable to contact ptdbnssd.\n");
	}
	ptdbnssd_timeout=1;
	close(udp_socket);
	return -1;
}
