#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <grp.h>
#include <netinet/in.h>
#include <nss.h>
#include <pthread.h>
#include <pwd.h>
#include <rx/rx.h>
#include <rx/xdr.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <afs/afsint.h>
// #include <afs/afs.h>
#include <afs/afsutil.h>
#include <afs/cellconfig.h>
#include <afs/com_err.h>
#include <afs/param.h>
#include <afs/ptclient.h>
#include <afs/pterror.h>
#include <afs/stds.h>
#include "ptdbnss.h"

extern struct ubik_client *pruclient;

int running=1;
int debug=0;
int udpport=PTDBNSSD_PORT;
int childsnum=NUMOFCHILDS;
char homedirs_method=0;
char shells_method=0;
pid_t *childs;
int pid;

/*
  Gets the username of a given uid
*/
int ptsid2name(uid_t uid, char *buffer, int buflen) {
	int code;
	idlist lid;
	namelist lnames;
	int i;
	
	lid.idlist_val = (afs_int32*)&uid;
	lid.idlist_len = 1;
	lnames.namelist_val = 0;
	lnames.namelist_len = 0;

	code = ubik_Call(PR_IDToName,pruclient,0,&lid,&lnames);
	if (code == PRSUCCESS) {
		for (i=0;i<lnames.namelist_len;i++) {
			int delta = strlen(lnames.namelist_val[i]);
			if ( delta < buflen && islower(*(lnames.namelist_val[i])) ) {
				strcpy(buffer, lnames.namelist_val[i]);
			}
		}
		free(lnames.namelist_val);
   
		/* free(lid.idlist_val); */
		lid.idlist_val = 0;
		lid.idlist_len = 0;
		return 1;
	} else {
		return -1;
	}
}




/*
  Gets the uid of a username
*/
uid_t ptsname2id( char *name)
 {
   int code;
   idlist lid;
   namelist lnames;
   int i;

   lid.idlist_val = 0;
   lid.idlist_len = 0;

   lnames.namelist_val = (prname*)name;
   lnames.namelist_len = 1;
   code = ubik_Call(PR_NameToID,pruclient,0,&lnames,&lid);
   if (code == PRSUCCESS) {
   	return (uid_t)lid.idlist_val[0];
   }
   return -1;
}

void init_afs() {
	int res;
	res=pr_Initialize(0L,AFSDIR_CLIENT_ETC_DIRPATH, 0);
	if ( debug ) {
		fprintf(stderr,"I: nssptdbd init done: res=%i, pid=%i\n",res,pid);
	}
}


/* Adds header version and configuration information to the answer-header */
void fill_header(struct ptdbnssheader_t *header) {
	header->version=HEADERVERSION;
	header->conf.homedirs_method=homedirs_method;
	header->conf.shells_method=shells_method;
}


/*
  This function will listen on udp://localhost:6998, processing
  one libnss-ptdb-request at a time. It's only executed in child
  processes.
*/
void serve(int udp_socket,struct sockaddr_in *laddr) {
	ssize_t ssize;
	ptdbnsspacket_t sendbuffer;
	socklen_t slen;
	struct sockaddr_in raddr;
	uid_t uid;
	int buflen;
	int res;
	pid=getpid();

	init_afs();

	while (1) {
		slen=sizeof(raddr);
		ssize=recvfrom(udp_socket,&sendbuffer,sizeof(ptdbnsspacket_t),0,(struct sockaddr*)&raddr,&slen);
		if ( sendbuffer.header.version == HEADERVERSION ) {
		// UID-to-Name-Lookup
		if (sendbuffer.header.ptype == NAME_REQUEST) {
			uid=sendbuffer.data.uid;
			if ( debug ) {
				fprintf(stderr,"D: pid %i: id2name(%i)...\n",pid,uid);
			}
			sendbuffer.data.name[0]=0;
			res=ptsid2name(uid,sendbuffer.data.name,ptdbnssdatalength-1);
			sendbuffer.header.ptype=NAME_REPLY;
			fill_header(&sendbuffer.header);
			sendto(udp_socket,&sendbuffer,sizeof(struct ptdbnssheader_t)+1+strlen(sendbuffer.data.name),0,(struct sockaddr*)&raddr,sizeof(raddr));
			if ( res == -1 ) {
				if ( debug ) {
					fprintf(stderr,"W: pid %i:  problem contacting PTDB, reinit'ing...\n",pid);
				}
				init_afs();
			} else {
				if ( debug ) {
					fprintf(stderr,"D: pid %i:  ='%s'\n",pid,sendbuffer.data.name);
				}
			}
		} else {
		
		// Name-to-UID-Lookup
		if (sendbuffer.header.ptype == UID_REQUEST) {
			// Make sure, there's a \0 at the end of the string
			sendbuffer.data.name[strnlen(sendbuffer.data.name,sizeof(ptdbnssdata_t)-1)+1]=0;
			if ( debug ) {
				fprintf(stderr,"D: pid %i: name2id(%s)...\n",pid,sendbuffer.data.name);
			}
			
			res=ptsname2id(sendbuffer.data.name);

			// AFS reports 32766 if the 
			if ( res == 32766 ) {
				if ( strcmp(sendbuffer.data.name,"anonymous") != 0 ) {
					if ( debug ) {
						fprintf(stderr,"D: pid %i:  =(invalid ID)\n",pid);
					}
					uid=0;
				} else {
					uid=32766;
					if ( debug ) {
						fprintf(stderr,"D: pid %i:  =32766\n",pid);
					}
				}		
			} else {
				uid=res;
				if ( debug ) {
					fprintf(stderr,"D: pid %i:  =%i\n",pid,uid);
				}
			}
			sendbuffer.data.uid=uid;
			sendbuffer.header.ptype=UID_REPLY;
			fill_header(&sendbuffer.header);
			sendto(udp_socket,&sendbuffer,sizeof(struct ptdbnssheader_t)+sizeof(sendbuffer.data.uid),0,(struct sockaddr*)&raddr,sizeof(raddr));
			if ( res == -1 ) {
				if ( debug ) {
					fprintf(stderr,"W: pid %i:  problem contacting PTDB, reinit'ing...\n",pid);
				}
				init_afs();
			}
		}
		}
		} else {
			if ( debug ) {
				fprintf(stderr,"E: pid %i: wrong header(%i)...\n",pid,sendbuffer.header.version);
			}
		}
	}
}


/*
	The signal handler for catching SIGTERM.
*/
void sig_term_handler(int signumber) {
	running=0;
	signal(SIGTERM,SIG_IGN);
	signal(SIGINT,SIG_IGN);
}


void help() {
	printf("Syntax: ptdbnssd {[options]}\n\n");
	printf("Valid options are:\n");
	printf(" -h, --help       this help screen\n");
	printf(" -d, --debug      debug mode\n");
	printf(" -p, --port=P     Bind to udp port P instead of %i\n",PTDBNSSD_PORT);
	printf(" -c, --childs=N   spawn N parallel server processes instead of %i\n",NUMOFCHILDS);
	printf(" -H, --homedirs=N Use method N to determine homedirs.\n");
	printf(" -S, --shells=N   Use method S to determine shells.\n\n");
	printf("Valid homedir methods:\n");
	printf(" user_1level       : /afs/$cell/user/$user\n");
	printf(" user_2level1x     : /afs/$cell/user/$user[0:1]/$user\n");
	printf(" user_3level12x    : /afs/$cell/user/$user[0:1]/$user[0:2]/$user\n");
	printf(" userext           : /afs/$cell/user-ext/$user\n");
	printf(" home_1level       : /afs/$cell/home/$user\n");
	printf(" admin             : Use homedir linked to by /afs/$cell/admin/homedirs/$user\n");
	printf(" local_home_minor  : Use /home/$user\n");
	printf(" local_home_capital: Use /home/$capital\n\n");
	printf("Valid shell methods:\n");
	printf(" bash            : /bin/bash\n");
	printf(" user            : Use binary linked to by symlink $homedir/.loginshell\n");
	printf(" admin           : Use binary linked to by symlink /afs/$cell/admin/shells/$user\n");
	printf(" scponly         : Use /usr/bin/scponly\n");
}


int main(int argc,char **argv,char **envp) {
	int childnum,childstatus;
	int udp_socket;
	struct sockaddr_in laddr;
	pid_t pid;
	int res;
	int c;
	int homedir_set=0;

	// Parameter parsing...
	while ( 1 ) {
		int option_index=0;
		static struct option long_options[] = {
			{"help",0,0,'h'},
			{"debug",0,0,'d'},
			{"port",1,0,'p'},
			{"childs",1,0,'c'},
			{"shells",1,0,'S'},
			{"homedirs",1,0,'H'},
			{0,0,0,0}
		};
		res=getopt_long (argc, argv, "dhp:c:S:H:", long_options, &option_index);
		if ( res == -1 ) break;
		switch ( res ) {
			case 'd':
				debug=1;
				break;
			case 'h':
				help();
				return 0;
				break;
			case 'p':
				udpport=atoi(optarg);
				break;
			case 'c':
				childsnum=atoi(optarg);
				break;
			case 'H':
				if ( ! strcmp("user_1level",optarg) ) {
					homedirs_method=HOMEDIR_USER_1LEVEL;
					homedir_set=1;
				}
				if ( ! strcmp("user_2level1x",optarg) ) {
					homedirs_method=HOMEDIR_USER_2LEVEL1X;
					homedir_set=1;
				}
				if ( ! strcmp("user_3level12x",optarg) ) {
					homedirs_method=HOMEDIR_USER_3LEVEL12X;
					homedir_set=1;
				}
				if ( ! strcmp("home_1level",optarg) ) {
					homedirs_method=HOMEDIR_HOME_1LEVEL;
					homedir_set=1;
				}
				if ( ! strcmp("userext",optarg) ) {
					homedirs_method=HOMEDIR_USEREXT;
					homedir_set=1;
				}
				if ( ! strcmp("auto",optarg) ) {
					homedirs_method=HOMEDIR_USER_1LEVEL;
					homedir_set=1;
				}
				if ( ! strcmp("admin",optarg) ) {
					homedirs_method=HOMEDIR_ADMINLINK;
					homedir_set=1;
				}
				if ( ! strcmp("local_home_minor",optarg) ) {
					homedirs_method=HOMEDIR_LOCAL_HOME_MINOR;
					homedir_set=1;
				}
				if ( ! strcmp("local_home_capital",optarg) ) {
					homedirs_method=HOMEDIR_LOCAL_HOME_CAPITAL;
					homedir_set=1;
				}
				if ( ! homedir_set ) {
					fprintf(stderr,"E: Unknown homedirectory method '%s'\n",optarg);
					exit(1);
				}
				if ( debug ) {
					fprintf(stderr,"D: setting homedirs_method to %i\n",homedirs_method);
				}
				break;
			case 'S':
				if ( ! strcmp("bash",optarg) ) {
					shells_method=SHELL_BASH;
				} else {
					if ( ! strcmp("user",optarg) ) {
						shells_method=SHELL_USERLINK;
					} else {
						if ( ! strcmp("user",optarg) ) {
							shells_method=SHELL_USERLINK;
						} else {
							if ( ! strcmp("scponly",optarg) ) {
								shells_method=SHELL_SCPONLY;
							} else {
								fprintf(stderr,"E: Unknown shell method '%s', ignored\n",optarg);
							}
						}
					}
				}
				if ( debug ) {
					fprintf(stderr,"D: setting shells_method to %i\n",shells_method);
				}
				break;
			default:
				fprintf(stderr,"getopt returned invalid result.\n");
		}
	}

	childs=malloc(sizeof(pid_t)*childsnum);
	if ( childs == NULL ) {
		fprintf(stderr,"E: Unable to malloc().\n");
		return 1;
	}
	
	// ptdbnssd uses udp://localhost:PTDBNSSD_PORT
	laddr.sin_family=AF_INET;
	laddr.sin_port=htons(udpport);
	laddr.sin_addr.s_addr=htonl(INADDR_LOOPBACK);
  
	udp_socket = socket(PF_INET, SOCK_DGRAM, 0);
	if ( bind(udp_socket,(struct sockaddr*)&laddr,sizeof(laddr)) == -1 ) {
		fprintf(stderr,"E: unable to bind to port udp://localhost:%i : %s\n",udpport,strerror(errno));
		return(0);
	}

	// multiple child processes will serve() in parallel
	for(c=0;c<childsnum;c++) {
		childs[c]=fork();
		if ( childs[c] == 0 ) {
			serve(udp_socket,&laddr);
		} else {
			if ( childs[c] == -1 ) {
				fprintf(stderr,"E: unable to fork()\n");
				return(1);
			}
		}
	}
	
	// Wait for a signal to come...
	signal(SIGTERM,sig_term_handler);
	signal(SIGINT,sig_term_handler);
	while ( running ) {
		sleep(1000);
	}
	for(c=0;c<childsnum;c++) {
		kill(childs[c],SIGTERM);
	}
	for(c=0;c<childsnum;c++) {
		wait(childstatus);
	}
	return(0);
}


